@@ -2,7 +2,7 @@
 #
 #  build-initrd.sh
 #
-#  Written by Jari Ruusu, June 17 2005
+#  Written by Jari Ruusu, September 17 2005
 #
 #  Copyright 2001-2005 by Jari Ruusu.
 #  Redistribution of this file is permitted under the GNU Public License.
@@ -129,7 +129,11 @@ INITRDONLY=0
 # Source root directory where files are copied from
 SOURCEROOT=
 
-# Destination root directory where files are written to
+# Destination root directory where files are written to.
+# Normally this is empty, but if you run this script on some other root
+# (i.e. Knoppix live CD), this must be configured to point to directory
+# where your about-to-be-encrypted root partition is mounted. This script
+# checks that an initrd directory exists there.
 DESTINATIONROOT=
 
 # dest-dir below dest-root
@@ -544,7 +548,7 @@ void _start()
         goto fail1;
     }
     if(pivot_root(".", "initrd")) {
-        wrStr("pivot_root() to new root failed. Older kernels don't have pivot_root()\n");
+        wrStr("pivot_root() to new root failed.\n Either 'initrd' directory is missing from your encrypted root partition\n or your kernel doesn't have pivot_root system call.\n");
         fail1:
         chdir("/");
         umount("/new-root");
