@@ -1,7 +1,7 @@
 /*
  *  glue.c
  *
- *  Written by Jari Ruusu, April 26 2005
+ *  Written by Jari Ruusu, October 20 2005
  *
  *  Copyright 2001-2005 by Jari Ruusu.
  *  Redistribution of this file is permitted under the GNU Public License.
@@ -30,6 +30,9 @@
 # include <linux/spinlock.h>
 #endif
 #include <asm/byteorder.h>
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+# include <asm/processor.h>
+#endif
 #include "aes.h"
 #include "md5.h"
 
@@ -60,7 +63,7 @@ typedef int TransferSector_t;
 # define LOOP_MULTI_KEY_SETUP_V3 0x4C4E
 #endif
 
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
 # define KEY_ALLOC_COUNT  128
 #else
 # define KEY_ALLOC_COUNT  64
@@ -69,7 +72,7 @@ typedef int TransferSector_t;
 typedef struct {
     aes_context *keyPtr[KEY_ALLOC_COUNT];
     unsigned    keyMask;
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     u_int32_t   *partialMD5;
     u_int32_t   partialMD5buf[8];
     rwlock_t    rwlock;
@@ -79,9 +82,37 @@ typedef struct {
 #else
     u_int32_t   partialMD5[4];
 #endif
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+    u_int32_t   padlock_cw_e;
+    u_int32_t   padlock_cw_d;
+#endif
 } AESmultiKey;
 
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+/* This function allocates AES context structures at special address such */
+/* that returned address % 16 == 8 . That way expanded encryption and */
+/* decryption keys in AES context structure are always 16 byte aligned */
+static void *specialAligned_kmalloc(size_t size, unsigned int flags)
+{
+    void *pn, **ps;
+    pn = kmalloc(size + (16 + 8), flags);
+    if(!pn) return (void *)0;
+    ps = (void **)((((unsigned long)pn + 15) & ~((unsigned long)15)) + 8);
+    *(ps - 1) = pn;
+    return (void *)ps;
+}
+static void specialAligned_kfree(void *ps)
+{
+    if(ps) kfree(*((void **)ps - 1));
+}
+# define specialAligned_ctxSize     ((sizeof(aes_context) + 15) & ~15)
+#else
+# define specialAligned_kmalloc     kmalloc
+# define specialAligned_kfree       kfree
+# define specialAligned_ctxSize     sizeof(aes_context)
+#endif
+
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
 static void keyScrubWork(AESmultiKey *m)
 {
     aes_context *a0, *a1;
@@ -116,7 +147,7 @@ static void keyScrubWork(AESmultiKey *m)
     p[3] ^= 0xFFFFFFFF;
 
     /* try to flush dirty cache data to RAM */
-#if defined(CONFIG_X86_64) || (defined(CONFIG_X86) && !defined(CONFIG_M386) && !defined(CONFIG_CPU_386))
+#if !defined(CONFIG_XEN) && (defined(CONFIG_X86_64) || (defined(CONFIG_X86) && !defined(CONFIG_M386) && !defined(CONFIG_CPU_386)))
     __asm__ __volatile__ ("wbinvd": : :"memory");
 #else
     mb();
@@ -169,20 +200,20 @@ static AESmultiKey *allocMultiKey(void)
     m = (AESmultiKey *) kmalloc(sizeof(AESmultiKey), GFP_KERNEL);
     if(!m) return 0;
     memset(m, 0, sizeof(AESmultiKey));
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     m->partialMD5 = &m->partialMD5buf[0];
     rwlock_init(&m->rwlock);
     init_timer(&m->timer);
     again:
 #endif
 
-    n = PAGE_SIZE / sizeof(aes_context);
+    n = PAGE_SIZE / specialAligned_ctxSize;
     if(!n) n = 1;
 
-    a = (aes_context *) kmalloc(sizeof(aes_context) * n, GFP_KERNEL);
+    a = (aes_context *) specialAligned_kmalloc(specialAligned_ctxSize * n, GFP_KERNEL);
     if(!a) {
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
-        if(x) kfree(m->keyPtr[0]);
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
+        if(x) specialAligned_kfree(m->keyPtr[0]);
 #endif
         kfree(m);
         return 0;
@@ -190,11 +221,11 @@ static AESmultiKey *allocMultiKey(void)
 
     while((x < KEY_ALLOC_COUNT) && n) {
         m->keyPtr[x] = a;
-        a++;
+        a = (aes_context *)((unsigned char *)a + specialAligned_ctxSize);
         x++;
         n--;
     }
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     if(x < 2) goto again;
 #endif
     return m;
@@ -205,21 +236,21 @@ static void clearAndFreeMultiKey(AESmultiKey *m)
     aes_context *a;
     int x, n;
 
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     /* stop scrub timer. loop thread was killed earlier */
     del_timer_sync(&m->timer);
     /* make sure allocated keys are in original order */
     if(m->reversed) keyScrubWork(m);
 #endif
-    n = PAGE_SIZE / sizeof(aes_context);
+    n = PAGE_SIZE / specialAligned_ctxSize;
     if(!n) n = 1;
 
     x = 0;
     while(x < KEY_ALLOC_COUNT) {
         a = m->keyPtr[x];
         if(!a) break;
-        memset(a, 0, sizeof(aes_context) * n);
-        kfree(a);
+        memset(a, 0, specialAligned_ctxSize * n);
+        specialAligned_kfree(a);
         x += n;
     }
 
@@ -245,7 +276,7 @@ static int multiKeySetup(struct loop_device *lo, unsigned char *k, int version3)
     m = (AESmultiKey *)lo->key_data;
     if(!m) return -ENXIO;
 
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     /* temporarily prevent loop thread from messing with keys */
     write_lock(&m->rwlock);
     m->blocked = 1;
@@ -253,13 +284,13 @@ static int multiKeySetup(struct loop_device *lo, unsigned char *k, int version3)
     if(m->reversed) keyScrubWork(m);
     write_unlock(&m->rwlock);
 #endif
-    n = PAGE_SIZE / sizeof(aes_context);
+    n = PAGE_SIZE / specialAligned_ctxSize;
     if(!n) n = 1;
 
     x = 0;
     while(x < KEY_ALLOC_COUNT) {
         if(!m->keyPtr[x]) {
-            a = (aes_context *) kmalloc(sizeof(aes_context) * n, GFP_KERNEL);
+            a = (aes_context *) specialAligned_kmalloc(specialAligned_ctxSize * n, GFP_KERNEL);
             if(!a) {
                 err = -ENOMEM;
                 goto error_out;
@@ -267,11 +298,11 @@ static int multiKeySetup(struct loop_device *lo, unsigned char *k, int version3)
             y = x;
             while((y < (x + n)) && (y < KEY_ALLOC_COUNT)) {
                 m->keyPtr[y] = a;
-                a++;
+                a = (aes_context *)((unsigned char *)a + specialAligned_ctxSize);
                 y++;
             }
         }
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
         if(x >= 64) {
             x++;
             continue;
@@ -311,7 +342,7 @@ static int multiKeySetup(struct loop_device *lo, unsigned char *k, int version3)
     lo->lo_flags |= 0x100000;   /* multi-key (info exported to user space) */
     memset(&un.b[0], 0, 32);
 error_out:
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     /* re-enable loop thread key scrubbing */
     write_lock(&m->rwlock);
     m->blocked = 0;
@@ -438,7 +469,7 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
     y = m->keyMask;
     if(cmd == READ) {
         while(size) {
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
             read_lock(&m->rwlock);
 #endif
             a = m->keyPtr[((unsigned)devSect) & y];
@@ -480,7 +511,7 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
                     loop_buf += 16;
                 }
             } while(--x >= 0);
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
             read_unlock(&m->rwlock);
 #endif
 #if LINUX_VERSION_CODE >= 0x20600
@@ -497,7 +528,7 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
         }
     } else {
         while(size) {
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
             read_lock(&m->rwlock);
 #endif
             a = m->keyPtr[((unsigned)devSect) & y];
@@ -561,7 +592,7 @@ int transfer_aes(struct loop_device *lo, int cmd, char *raw_buf,
                     raw_buf += 16;
                 } while(--x >= 0);
             }
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
             read_unlock(&m->rwlock);
 #endif
 #if LINUX_VERSION_CODE >= 0x20600
@@ -594,7 +625,31 @@ int keySetup_aes(struct loop_device *lo, LoopInfo_t *info)
     aes_set_key(m->keyPtr[0], &un.b[0], info->lo_encrypt_key_size, 0);
     memset(&info->lo_encrypt_key[0], 0, sizeof(info->lo_encrypt_key));
     memset(&un.b[0], 0, 32);
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+    switch(info->lo_encrypt_key_size) {
+    case 256:   /* bits */
+    case 32:    /* bytes */
+        /* 14 rounds, AES, software key gen, normal oper, encrypt, 256-bit key */
+        m->padlock_cw_e = 14 | (1<<7) | (2<<10);
+        /* 14 rounds, AES, software key gen, normal oper, decrypt, 256-bit key */
+        m->padlock_cw_d = 14 | (1<<7) | (1<<9) | (2<<10);
+        break;
+    case 192:   /* bits */
+    case 24:    /* bytes */
+        /* 12 rounds, AES, software key gen, normal oper, encrypt, 192-bit key */
+        m->padlock_cw_e = 12 | (1<<7) | (1<<10);
+        /* 12 rounds, AES, software key gen, normal oper, decrypt, 192-bit key */
+        m->padlock_cw_d = 12 | (1<<7) | (1<<9) | (1<<10);
+        break;
+    default:
+        /* 10 rounds, AES, software key gen, normal oper, encrypt, 128-bit key */
+        m->padlock_cw_e = 10 | (1<<7);
+        /* 10 rounds, AES, software key gen, normal oper, decrypt, 128-bit key */
+        m->padlock_cw_d = 10 | (1<<7) | (1<<9);
+        break;
+    }
+#endif
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     keyScrubTimerInit(lo);
 #endif
     return(0);
@@ -609,6 +664,115 @@ int keyClean_aes(struct loop_device *lo)
     return(0);
 }
 
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+#if LINUX_VERSION_CODE < 0x20400
+#error "this code does not support padlock crypto instructions on 2.2 or older kernels"
+#endif
+
+static __inline__ void padlock_flush_key_context(void)
+{
+    __asm__ __volatile__("pushf; popf" : : : "cc");
+}
+
+static __inline__ void padlock_rep_xcryptcbc(void *cw, void *k, void *s, void *d, void *iv, unsigned long cnt)
+{
+    __asm__ __volatile__(".byte 0xF3,0x0F,0xA7,0xD0"
+                         : "+a" (iv), "+c" (cnt), "+S" (s), "+D" (d) /*output*/
+                         : "b" (k), "d" (cw) /*input*/
+                         : "cc", "memory" /*modified*/ );
+}
+
+typedef struct {
+    u_int32_t   iv[4];
+    u_int32_t   cw[4];
+    u_int32_t   dummy1[4];
+} Padlock_IV_CW;
+
+static int transfer_padlock_aes(struct loop_device *lo, int cmd, char *raw_buf,
+          char *loop_buf, int size, TransferSector_t devSect)
+{
+    aes_context     *a;
+    AESmultiKey     *m;
+    unsigned        y;
+    Padlock_IV_CW   ivcwua;
+    Padlock_IV_CW   *ivcw;
+
+    /* ivcw->iv and ivcw->cw must have 16 byte alignment */
+    ivcw = (Padlock_IV_CW *)(((unsigned long)&ivcwua + 15) & ~((unsigned long)15));
+
+    if(!size || (size & 511) || (((unsigned long)raw_buf | (unsigned long)loop_buf) & 15)) {
+        return -EINVAL;
+    }
+    m = (AESmultiKey *)lo->key_data;
+    y = m->keyMask;
+    if(cmd == READ) {
+        while(size) {
+            padlock_flush_key_context();
+            ivcw->cw[0] = m->padlock_cw_d;
+            ivcw->cw[3] = ivcw->cw[2] = ivcw->cw[1] = 0;
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
+            read_lock(&m->rwlock);
+#endif
+            a = m->keyPtr[((unsigned)devSect) & y];
+            if(y) {
+                memcpy(&ivcw->iv[0], raw_buf, 16);
+                padlock_rep_xcryptcbc(&ivcw->cw[0], &a->aes_d_key[0], raw_buf + 16, loop_buf + 16, &ivcw->iv[0], 31);
+                memcpy(&ivcw->iv[0], &m->partialMD5[0], 16);
+                loop_compute_md5_iv_v3(devSect, &ivcw->iv[0], (u_int32_t *)(&loop_buf[16]));
+                padlock_rep_xcryptcbc(&ivcw->cw[0], &a->aes_d_key[0], raw_buf, loop_buf, &ivcw->iv[0], 1);
+            } else {
+                loop_compute_sector_iv(devSect, &ivcw->iv[0]);
+                padlock_rep_xcryptcbc(&ivcw->cw[0], &a->aes_d_key[0], raw_buf, loop_buf, &ivcw->iv[0], 32);
+            }
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
+            read_unlock(&m->rwlock);
+#endif
+#if LINUX_VERSION_CODE >= 0x20600
+            cond_resched();
+#else
+            if(current->need_resched) {set_current_state(TASK_RUNNING);schedule();}
+#endif
+            size -= 512;
+            raw_buf += 512;
+            loop_buf += 512;
+            devSect++;
+        }
+    } else {
+        while(size) {
+            padlock_flush_key_context();
+            ivcw->cw[0] = m->padlock_cw_e;
+            ivcw->cw[3] = ivcw->cw[2] = ivcw->cw[1] = 0;
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
+            read_lock(&m->rwlock);
+#endif
+            a = m->keyPtr[((unsigned)devSect) & y];
+            if(y) {
+                memcpy(raw_buf, loop_buf, 512);
+                memcpy(&ivcw->iv[0], &m->partialMD5[0], 16);
+                loop_compute_md5_iv_v3(devSect, &ivcw->iv[0], (u_int32_t *)(&raw_buf[16]));
+                padlock_rep_xcryptcbc(&ivcw->cw[0], &a->aes_e_key[0], raw_buf, raw_buf, &ivcw->iv[0], 32);
+            } else {
+                loop_compute_sector_iv(devSect, &ivcw->iv[0]);
+                padlock_rep_xcryptcbc(&ivcw->cw[0], &a->aes_e_key[0], loop_buf, raw_buf, &ivcw->iv[0], 32);
+            }
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
+            read_unlock(&m->rwlock);
+#endif
+#if LINUX_VERSION_CODE >= 0x20600
+            cond_resched();
+#else
+            if(current->need_resched) {set_current_state(TASK_RUNNING);schedule();}
+#endif
+            size -= 512;
+            raw_buf += 512;
+            loop_buf += 512;
+            devSect++;
+        }
+    }
+    return(0);
+}
+#endif
+
 int handleIoctl_aes(struct loop_device *lo, int cmd, unsigned long arg)
 {
     int err;
@@ -636,13 +800,33 @@ static struct loop_func_table funcs_aes = {
     ioctl:      (void *) handleIoctl_aes
 };
 
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+static struct loop_func_table funcs_padlock_aes = {
+    number:     16,     /* 16 == AES */
+    transfer:   (void *) transfer_padlock_aes,
+    init:       (void *) keySetup_aes,
+    release:    keyClean_aes,
+    ioctl:      (void *) handleIoctl_aes
+};
+#endif
+
 int init_module_aes(void)
 {
+#if defined(CONFIG_BLK_DEV_LOOP_PADLOCK) && (defined(CONFIG_X86) || defined(CONFIG_X86_64))
+    if((boot_cpu_data.x86 >= 6) && (boot_cpu_data.x86_vendor == X86_VENDOR_CENTAUR)
+      && (cpuid_eax(0xC0000000) >= 0xC0000001) && ((cpuid_edx(0xC0000001) & 0xC0) == 0xC0)) {
+        if(loop_register_transfer(&funcs_padlock_aes)) {
+            printk("loop: unable to register padlock AES transfer\n");
+            return -EIO;
+        }
+        printk("loop: padlock hardware AES enabled\n");
+    } else
+#endif
     if(loop_register_transfer(&funcs_aes)) {
         printk("loop: unable to register AES transfer\n");
         return -EIO;
     }
-#if CONFIG_BLK_DEV_LOOP_KEYSCRUB
+#ifdef CONFIG_BLK_DEV_LOOP_KEYSCRUB
     printk("loop: AES key scrubbing enabled\n");
 #endif
     return 0;
